/*
**   line based config file parsing library -- csconfig.c
**   Copyright (C) 2001-2004 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>
#include "csconfig.h"


typedef struct _csc_info		CSC_INFO;

struct _csc_info
{
	char				*csci_filename;
	u_int				csci_lineno;

	char				*csci_curptr;
	char				csci_save;
	char				csci_recurs;

#define CSC_CSCI_FLAGS_INBLOCK		0x001
	u_int				csci_flags;

	csc_error_cb_t		csci_error_cb;

	CSC_COMMAND			*csci_command;

};

static int _csc_parse(CSC_INFO *csci, CSC_COMMAND *commands, void *cb_arg);


static int _csc_parse(CSC_INFO *csci, CSC_COMMAND *commands, void *cb_arg)
{
	register CSC_COMMAND	*comptr;
	register u_int			flags = csci->csci_flags;
	register char			*ptr = csci->csci_curptr;
	void					*cb_arg2;
	char					*command;
	char					*start;
	int						err;

	for(;csci->csci_save;*ptr++ = csci->csci_save)
	{
		start = ptr;
		while(*ptr && (*ptr != '\n'))
		{
			++ptr;
			continue;
		}
		csci->csci_save = *ptr;
		*ptr = '\0';
		++csci->csci_lineno;

		if (!*start || (*start == '#'))
		{
			continue;
		}

		command = csconfig_next_arg(&start);
		if (!command || !*command)
		{
			continue;
		}

		for(comptr=commands;comptr->cscc_name;comptr++)
		{
			if ((flags & CSC_CSCI_FLAGS_INBLOCK) &&
					!strcasecmp(command, "end"))
			{
				csci->csci_curptr = ptr;
				return 0;
			}
			if (strcasecmp(command, comptr->cscc_name))
				continue;
			csci->csci_command = comptr;
			cb_arg2 = cb_arg;
			err = comptr->cscc_cb((void *)csci, start, &cb_arg2);
			if (err)
				return err;
			if (comptr->cscc_arg)
			{
				csci->csci_flags |= CSC_CSCI_FLAGS_INBLOCK;
				*ptr++ = csci->csci_save;
				csci->csci_curptr = ptr;
				csci->csci_recurs++;
				err = _csc_parse(csci, (CSC_COMMAND *)comptr->cscc_arg, cb_arg2);
				if (err)
					return err;
				csci->csci_recurs--;
				ptr = csci->csci_curptr;
				if (!(flags & CSC_CSCI_FLAGS_INBLOCK))
				{
					csci->csci_flags &= ~CSC_CSCI_FLAGS_INBLOCK;
				}
			}
			break;
		}
		if (!comptr->cscc_name)
		{
			if (csci->csci_error_cb)
				csci->csci_error_cb(CSC_SEVERITY_WARNING, "Unknown command",
						csci->csci_filename, csci->csci_lineno);
		}
	}
	if (flags & CSC_CSCI_FLAGS_INBLOCK)
	{
		if (csci->csci_error_cb)
			csci->csci_error_cb(CSC_SEVERITY_WARNING, "No 'end' of block",
					csci->csci_filename, csci->csci_lineno);
	}
	csci->csci_curptr = ptr;
	return 0;
}

char *csconfig_next_arg(char **args)
{
	char	*ptr;
	char	*start;

	if (!args || !*args)
		return NULL;
	ptr = *args;
	while((*ptr == ' ') || (*ptr == '\t'))
		ptr++;
	if (!*ptr)
	{
		*args = NULL;
		return ptr;
	}
	start = ptr;
	while(*ptr && ((*ptr != ' ') && (*ptr != '\t')))
		ptr++;
	if (!*ptr)
		*args = NULL;
	else
	{
		*ptr++ = '\0';
		*args = ptr;
	}
	return start;
}

/*
** Returns 0 on success
**   return value > 0 means its an errno value.
**   return value < 0 means there was some other error.
*/
int csconfig_parse(char *buffer, CSC_COMMAND *commands, void *cb_arg, csc_error_cb_t error_cb)
{
	CSC_INFO		cscinfo;
	CSC_INFO		*csci = &cscinfo;

	if (!buffer || !commands)
	{
		return EINVAL;
	}

	csci->csci_filename = "<buffer>";
	csci->csci_lineno = 0;
	csci->csci_recurs = 0;
	csci->csci_save = ' ';
	csci->csci_curptr = buffer;
	csci->csci_flags = 0;
	csci->csci_command = NULL;
	csci->csci_error_cb = error_cb;

	return _csc_parse(csci, commands, cb_arg);
}

int csconfig_read(char *filename, CSC_COMMAND *commands, void *cb_arg, csc_error_cb_t error_cb)
{
	int				fd;
	int				err;
	struct stat		sb;
	char			*cbuf;
	CSC_INFO		cscinfo;
	CSC_INFO		*csci = &cscinfo;

	if (!filename || !commands)
	{
		return EINVAL;
	}

	fd = open(filename, O_RDONLY);
	if (fd < 0)
	{
		return errno ? errno : -1;
	}

	if (fstat(fd, &sb) < 0)
	{
		err = errno ? errno : -1;
		close(fd);
		return err;
	}

	if (!sb.st_size)
	{
		close(fd);
		return 0;
	}

	cbuf = (char *)malloc(sb.st_size + 1);
	if (cbuf == NULL)
	{
		err = errno ? errno : ENOMEM;
		close(fd);
		return err;
	}

	err = read(fd, cbuf, sb.st_size);
	if (err < 0)
	{
		err = errno ? errno : -1;
		close(fd);
		free(cbuf);
		return err;
	}
	else if (err != sb.st_size)
	{
		close(fd);
		free(cbuf);
		return -1;
	}
	close(fd);
	cbuf[err] = '\0';

	csci->csci_filename = filename;
	csci->csci_lineno = 0;
	csci->csci_recurs = 0;
	csci->csci_save = ' ';
	csci->csci_curptr = cbuf;
	csci->csci_flags = 0;
	csci->csci_command = NULL;
	csci->csci_error_cb = error_cb;

	err = _csc_parse(csci, commands, cb_arg);
	free(cbuf);

	return err;
}

u_int csconfig_lineno(void *csc_cookie)
{
	return ((CSC_INFO *)csc_cookie)->csci_lineno;
}

char *csconfig_filename(void *csc_cookie)
{
	return ((CSC_INFO *)csc_cookie)->csci_filename;
}

CSC_COMMAND *csconfig_command(void *csc_cookie)
{
	return ((CSC_INFO *)csc_cookie)->csci_command;
}

void csconfig_error(void *csc_cookie, int severity, char *string)
{
	CSC_INFO	*csci = (CSC_INFO *)csc_cookie;

	if (string && csci && csci->csci_error_cb)
		csci->csci_error_cb(severity, string,
			csci->csci_filename, csci->csci_lineno);
}
